package ch.usi.inf.sape.tracer.analyzer;

/**
 * This class provides a total ordering among the events executed
 * in different threads of execution. 
 * 
 * Each thread has its own event counter. However, we need a total
 * order among the events executed in different threads. This class
 * is intended to provide this ordering.
 * 
 * @author reza
 *
 */
public final class GlobalClock {
	
	private static long time = 1;
	
	public synchronized static long getAndIncrementTime(){
		if (time == Long.MAX_VALUE)
			throw new IllegalStateException("Global time counter overflowing...!");
		
		return time++;
	}
}
