package ch.usi.inf.sape.tracer.analyzer;

import java.io.BufferedOutputStream;
import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Iterator;
import java.util.List;

import ch.usi.inf.sape.tracer.analyzer.Activation.UnTracedActivation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.ANEWARRAY_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.ARRAYLENGTH_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.ATHROW_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BIPUSH_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BasicBlockEntry_PseudoOperation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.Binary_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.CHECKCAST_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.CatchBlock_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUMMY_RETURN_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_1_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_2_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_X1_1_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_X1_2_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_X2_1_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_X2_2_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_X2_3_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_X2_4_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP_X1_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP_X2_1_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP_X2_2_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.Division_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.GETFIELD_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.GETSTATIC_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.GOTO_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.IF_ACMP_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.IF_ICMP_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.IF_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.IINC_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.INSTANCEOF_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.INVOKEARGS_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.INVOKEARGS_Operation.UnTracedInvokeArgs_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.INVOKECALLING_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.INVOKERETURNING_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.LOOKUPSWITCH_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.MONITORENTER_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.MONITOREXIT_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.MULTIANEWARRAY_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.NEWARRAY_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.NEW_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.PUTFIELD_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.PUTSTATIC_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.Rem_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.SIPUSH_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.SWAP_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.TABLESWITCH_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.Unary_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.UntracedBytecode_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XALOAD_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XASTORE_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XCONST_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XLDC_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XLOAD_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XPOP_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XRETURN_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XSTORE_Operation;
import ch.usi.inf.sape.tracer.analyzer.locations.ArrayElementMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.ArrayLengthMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.ArrayMetaData;
import ch.usi.inf.sape.tracer.analyzer.locations.ClassFieldMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.ClassMetaData;
import ch.usi.inf.sape.tracer.analyzer.locations.InstanceFieldMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.InstanceMetaData;
import ch.usi.inf.sape.tracer.analyzer.locations.LocalMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.PseudoVariableMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.ReturnValueMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.StackSlotMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.TemporaryMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.UndefinedMemoryLocation;

public class JSONWriterVisitor implements TraceElementVisitor{
	
	private final PrintWriter out;
	private int indent;
	private long time;
	
	public JSONWriterVisitor(String filename) throws FileNotFoundException {
		out = new PrintWriter(new BufferedOutputStream(new FileOutputStream(filename)));
	}
	
	
	
//	public void finalizeJsonFile(){
//		indent();
//		out.println("]");
//		
//		indent--;
//		out.println("}");
//		out.close();
//	}
	
	private void indent() {
        for (int i=0; i<indent; i++) {
            out.print("  ");
        }
    }


    public void visitGenerally(BytecodeEvent e) throws VisitorException{
    	setTime(e.getEventTime());
    	indent();
    	out.println("\"name\": \""+ e.getOperationName()+"\",");
    	indent();
    	out.println("\"time\": "+e.getEventTime()+",");
    	indent();
    	out.println("\"source-name\": \""+e.getSourceFileName()+"\",");
    	indent();
    	out.println("\"source-line-no\": "+e.getSourceCodeLineNumber()+",");
    	indent();
    	out.println("\"activation\": {");
    	indent++;
    	visit(e.getActivationRecord());
    	indent--;
    	indent();
    	out.println("}");
    }
    
	@Override
	public Object visit(ANEWARRAY_Operation anewarray_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(ARRAYLENGTH_Operation arraylength_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(ATHROW_Operation athrow_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(BasicBlockEntry_PseudoOperation basicBlockEntry_PseudoOperation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(Binary_Operation binary_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(Division_Operation division_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(Rem_Operation rem_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(BIPUSH_Operation bipush_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(CatchBlock_Operation catchBlock_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(CHECKCAST_Operation checkcast_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(DUMMY_RETURN_Operation dummy_RETURN_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(DUP_Operation dup_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(DUP_X1_Operation dup_X1_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(DUP_X2_1_Operation dup_X2_1_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(DUP_X2_2_Operation dup_X2_2_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(DUP2_1_Operation dup2_1_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(DUP2_2_Operation dup2_2_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(DUP2_X1_1_Operation dup2_X1_1_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(DUP2_X1_2_Operation dup2_X1_2_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(DUP2_X2_1_Operation dup2_X2_1_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(DUP2_X2_2_Operation dup2_X2_2_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(DUP2_X2_3_Operation dup2_X2_3_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(DUP2_X2_4_Operation dup2_X2_4_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(GETFIELD_Operation getfield_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(GETSTATIC_Operation getstatic_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(GOTO_Operation goto_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(IF_ACMP_Operation if_ACMP_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(IF_ICMP_Operation if_ICMP_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(IF_Operation if_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(IINC_Operation iinc_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(INSTANCEOF_Operation instanceof_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(INVOKEARGS_Operation o) throws VisitorException {
		
		return null;
	}

	@Override
	public Object visit(UnTracedInvokeArgs_Operation o) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(INVOKECALLING_Operation invokecalling_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(INVOKERETURNING_Operation invokereturning_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(LOOKUPSWITCH_Operation lookupswitch_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(MONITORENTER_Operation monitorenter_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(MONITOREXIT_Operation monitorexit_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(MULTIANEWARRAY_Operation multianewarray_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(NEW_Operation new_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(NEWARRAY_Operation newarray_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(PUTFIELD_Operation e) throws VisitorException {
		Binding<InstanceFieldMemoryLocation> binding = (Binding<InstanceFieldMemoryLocation>) e.getDefs()[0];
		InstanceMetaData instance = binding.getMemoryLocation().getContainer();
		Binding<? extends MemoryLocation>[] elements = instance.getElementsAtTime(e.getEventTime());
		
		indent();
		out.println(",");
		
		indent();
		out.println("\"object\": {");
		indent++;
		
		indent();
		out.println("\"oid\": \""+instance.getId()+"\",");
		
		indent();
		out.println("\"updated-field-name\": \""+binding.getMemoryLocation().getName()+"\",");
		
		indent();
		out.println("\"fields\": [");
		indent++;
		
		boolean first = true;
		for (Binding<? extends MemoryLocation> ebinding : elements) {
			if (!first){
				indent();
				out.println(",");
			}
			indent();
			out.println("{");
			first = false;
			
			Binding<InstanceFieldMemoryLocation> insBinding = (Binding<InstanceFieldMemoryLocation>) ebinding;
			indent();
			out.println("\"name\": \""+ insBinding.getMemoryLocation().getName()+"\",");
			
			indent();
			out.println("\"value\": \""+ebinding.getValue().getValueString()+"\"");
			
			indent();
			out.println("}");
		}
		
		indent--;
		indent();
		out.println("]");
		
		indent--;
		indent();
		out.println("}");
		return null;
	}

	@Override
	public Object visit(PUTSTATIC_Operation putstatic_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(SIPUSH_Operation sipush_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(SWAP_Operation swap_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(TABLESWITCH_Operation tableswitch_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(Unary_Operation unary_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(UntracedBytecode_Operation untracedBytecode_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(XALOAD_Operation xaload_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(XASTORE_Operation xastore_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(XCONST_Operation xconst_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(XLDC_Operation xldc_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(XLOAD_Operation xload_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(XPOP_Operation xpop_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(XRETURN_Operation xreturn_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(XSTORE_Operation xstore_Operation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(ArrayElementMemoryLocation arrayElementMemoryLocation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(ArrayLengthMemoryLocation arrayLengthMemoryLocation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(ClassFieldMemoryLocation classFieldMemoryLocation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(InstanceFieldMemoryLocation instanceFieldMemoryLocation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(LocalMemoryLocation localMemoryLocation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(UndefinedMemoryLocation undefinedMemoryLocation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(PseudoVariableMemoryLocation pseudoVariableMemoryLocation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(ReturnValueMemoryLocation returnValueMemoryLocation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(StackSlotMemoryLocation stackSlotMemoryLocation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(TemporaryMemoryLocation temporaryMemoryLocation) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(Activation a) throws VisitorException {
		indent();
		out.println("\"name\": \""+a.getMethodName()+"\",");
		
		indent();
		out.println("\"locals\": [");
		indent++;
		Binding<? extends MemoryLocation>[] bs = a.getElementsAtTime(getTime());
		
		boolean first = true;
		for (Binding<? extends MemoryLocation> binding : bs) {
			
			if (binding == Binding.UNDEFINED_BINDING)
				continue;
				
			if (!first){
				indent();
				out.println(",");
			}
			
			first = false;
			
			LocalMemoryLocation l = (LocalMemoryLocation) binding.getMemoryLocation();
			indent();
			out.println("{");
			
			indent();
			out.println("\"name\": \""+l.getName()+"\",");
			
			indent();
			out.println("\"value\": \""+binding.getValue().getValueString()+"\"");
			
			indent();
			out.println("}");
			
			
		}
		indent--;
		indent();
		out.println("]");
		return null;
	}

	@Override
	public Object visit(UnTracedActivation a) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(ArrayMetaData arrayMetaData) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(ClassMetaData classMetaData) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(InstanceMetaData instanceMetaData) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(ThreadCallStack threadCallStack) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(Value v) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(ArrayReferenceValue av) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(InstanceReferenceValue iv) throws VisitorException {
		return null;
	}

	@Override
	public Object visit(Binding<?> b) throws VisitorException {
		return null;
	}
	
	public void scan(List<BytecodeEvent> events) throws VisitorException{
		out.println("[");
		
		indent++;
		Iterator<BytecodeEvent> it = events.iterator();
		while(it.hasNext()){
			indent();
			out.println("{");
			
			indent++;
			
			BytecodeEvent e = it.next();
			visitGenerally(e);
			e.accept(this);
			
			indent--;
			
			indent();
			out.println("}");
			
			
			if (it.hasNext()){
				indent();
				out.println(",");
			}
		}
		indent--;
		
		indent();
		out.println("]");
		out.close();
	}

	
    /**
	 * @return the time
	 */
	public long getTime() {
		return time;
	}



	/**
	 * @param time the time to set
	 */
	public void setTime(long time) {
		this.time = time;
	}

}
