package ch.usi.inf.sape.tracer.analyzer;

import java.util.List;

import org.junit.runner.JUnitCore;
import org.junit.runner.Request;
import org.junit.runner.Result;
import org.junit.runner.notification.Failure;

public class JUnitTestDriver {

	private final String testClassName;

	private final String testMethodName;

	private final ClassLoader appCL;

	public JUnitTestDriver(String testClassName, String testMethodName, ClassLoader appCL) {
		this.testClassName = testClassName;
		this.testMethodName = testMethodName;
		this.appCL = appCL;
	}


	public void launch(){
		Result result = null;
		try {
			System.out.println("JUnitTestWrapper Run Test:\t"+ testClassName+"."+testMethodName);
			Class<?> klass = appCL.loadClass(testClassName);
			final Request request = Request.method(klass, testMethodName);
			final JUnitCore junit = new JUnitCore();
			result = junit.run(request);

			System.out.println(getFailureMessages(result));
		} catch(Throwable e){
			e.printStackTrace(System.out);
		} finally{
			System.out.println("Test run [done]");
			synchronized (Thread.currentThread()){
				Thread.currentThread().notifyAll();
			}
		}

	}
	
	
	private String getFailureMessages(Result result){
		List<Failure> list =result.getFailures();
		String messages = "";
		for (Failure failure : list) {
			messages += failure.getTrace()+"\n";
		}
		return messages;
	}
}
