package ch.usi.inf.sape.tracer.analyzer;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import ch.usi.inf.sape.tracer.analyzer.abstractops.SourceLineOperation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.RuntimeEnv;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.SlicedBytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.locations.LocalMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.PseudoVariableMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.StackSlotMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocation.LocationType;

public class QueryFacade {

	public static boolean isInvokeInstruction(Event op){
		int opCode = op.getOperationCode();
		if ( opCode == Bytecodes.INVOKEINTERFACE ||
				opCode == Bytecodes.INVOKESPECIAL ||
				opCode == Bytecodes.INVOKESTATIC ||
				opCode == Bytecodes.INVOKEVIRTUAL)
			return true;
		
		return false;
	}
	
	public static Activation getActivationRecordFor(SlicedBytecodeEvent op, RuntimeEnv session){
		return getActivationRecordFor(op.getNonSlicedEvent(), session);
	}
	
	public static Activation getActivationRecordFor(BytecodeEvent op, RuntimeEnv session){
		return (Activation) findControllerForBytecode(op, session).getDefs()[0].getMemoryLocation().getContainer();
	}
	
	public static SourceLineOperation getNthOccurenceOfSourceLineNo(int nth, int sourceLineNo, RuntimeEnv env){
		List<BytecodeEvent> bcs = env.getAllBytecodeEvents();
		List<BytecodeEvent> occurences = new ArrayList<>();
		int index = 0;
		int match = 1;
		while (index < bcs.size()){
			if (bcs.get(index).getSourceCodeLineNumber() != sourceLineNo){
				index ++;
				continue;
			}else if (match == nth){
				do{
					occurences.add(bcs.get(index));
					index++;
				}while(bcs.get(index).getSourceCodeLineNumber() == sourceLineNo || 
						bcs.get(index).getSourceCodeLineNumber() == SourceLineOperation.PSEUDO_VAR_OPERATION_LINE_NO);
				break;
			}else {
				do {
					index ++;
				}while(bcs.get(index).getSourceCodeLineNumber() == sourceLineNo);
				match ++;
			}
		}
		if (occurences.size() == 0){
			throw new IllegalArgumentException("Line number "+ sourceLineNo+ " never occured in this run!");
		}
		
		return new SourceLineOperation(occurences.toArray(new BytecodeEvent[0]), occurences.get(0).getOperationIndex(), sourceLineNo, nth);
	}
	
//	public static String findStackTrace(List<? extends Operation> ops){
//		final List<Activation> frames = new ArrayList<Activation>();
//		final List<Integer> lineNos = new ArrayList<Integer>();
//		
//		for (Operation operation : ops){
//			if (isInvokeInstruction(operation)){
//				frames.add(((INVOKEARGS_Operation)operation.getNonSlicedOperation()).getCalleeFrame());
//				lineNos.add(operation.getSourceCodeLineNumber());
//			}
//		}
//		
//		final StringBuilder builder = new StringBuilder();
//		for (int i = 0; i < frames.size(); i++) {
//			builder.append("\tat ").append(frames.get(i).getOwnerClass().replace('/', '.'))
//							.append(".")
//							.append(frames.get(i).getMethodName())
//							.append(frames.get(i).getDesc())
//							.append("(").append("Line: ").append(lineNos.get(i)).append(")\n");
//		} 
//		
//		return builder.toString();
//	}
	
	public static String stack(Binding<? extends MemoryLocation> l){
		BytecodeEvent bc = l.getProvider();
		Activation container = (Activation)l.getMemoryLocation().getContainer();
		
		final StringBuilder builder = new StringBuilder();
		while(true){
			System.out.println(bc);
			int sourceLineNo = bc.getSourceCodeLineNumber();
			if (container == null || container.getProvider() == null)
				break;
			
			builder.append("\tat ").append(container.getOwnerClass().replace('/', '.'))
			.append(".")
			.append(container.getMethodName())
			.append(container.getDesc())
			.append("(").append("Line: ").append(sourceLineNo).append(")\n");
			
			bc = container.getProvider();
			container = container.getProvider().getCallerFrame();
		}
		return builder.toString();
	}


	public static boolean isReturnInstruction(BytecodeEvent op) {
		// TODO Auto-generated method stub
		return false;
	}


	public static boolean isObjectAllocation(BytecodeEvent op) {
		// TODO Auto-generated method stub
		return false;
	}
	
	
	public static EventList<BytecodeEvent> getWriteIntoArrayElementEvents(RuntimeEnv env){
		List<BytecodeEvent> writes = new ArrayList<>();
		List<BytecodeEvent> events = env.getAllBytecodeEvents();
		for (BytecodeEvent event : events) {
			for (Binding<?> use : event.getDefs()){
				if (use.getMemoryLocation().getLocationType() == LocationType.ARRAY_ELEMENT)
					writes.add(event);
			}
		}
		return new EventList<BytecodeEvent>(writes);
	}
	
	
	

	/* *********************************************************************************
	 ******************************* Slicing Operations ********************************
	 ***********************************************************************************/
		
	/**
	 * This is a full backward slice including both data dependency and control dependency.
	 * 
	 * @param op
	 * @return
	 */
	public static List<SlicedBytecodeEvent> findBackwardSliceForBytecode(final BytecodeEvent criterion, 
												 							final boolean includeData, 
												 							final boolean includeControl, 
												 							final RuntimeEnv env){
		Set<SlicedBytecodeEvent> backwardSliceSet = new HashSet<>();
	
		/*
		 	Perform data flow analysis first. For data flow analysis, for every bytecode, we find
		 	only those origins that have been used for defining our considered target value, i.e.,
		 	we use the getUsesForDef() method.
		 */
		int sessionId = env.getAnalysisSession().getAndIncrementSessionId();
		SlicedBytecodeEvent slicedOp = criterion.getSlicedOperationForSession(sessionId);
		slicedOp.addUses(criterion.getUses());		// for the original slice criterion, we add all its uses to the use list of the sliced operation
		backwardSliceSet.add(slicedOp);				// add the slice criterion 
		if (includeData){
			for (Binding<? extends MemoryLocation> use : criterion.getUses()){
				backwardSliceSet.addAll(findBackwardSliceForBytecode(use.getProvider(), 
																		use, 
																		sessionId, 
																		includeData, 
																		includeControl, 
																		env));
			}
		}
		
		/*
		 	In the second step, we perform control flow analysis  
		 */
		if (includeControl){
			backwardSliceSet.addAll(findBackwardSliceForBytecode(criterion, sessionId, includeData, includeControl, env));
		}
		
		List<SlicedBytecodeEvent> backwardSliceList = new ArrayList<>(backwardSliceSet);
		Collections.sort(backwardSliceList, new Comparator<SlicedEvent>() {

			@Override
			public int compare(SlicedEvent o1, SlicedEvent o2) {
				return o1.getEventTime() < o2.getEventTime() ? -1 : o1.getEventTime() == o2.getEventTime() ? 0 : 1;
			}
		});
		
		return backwardSliceList;
	}
	
	private static Set<SlicedBytecodeEvent> findBackwardSliceForBytecode(final BytecodeEvent criterion, 
																			Binding<? extends MemoryLocation> def, 
																			int sessionId,
																			boolean includeData, 
																			boolean includeControl,
																			RuntimeEnv env){
		Set<SlicedBytecodeEvent> ops = new HashSet<>();
		if (criterion == BytecodeEvent.UNTRACED_BYTECODE_OPERATION){
			return ops;
		}
		//Data flow analysis  
		if (includeData){
			SlicedBytecodeEvent slicedOp = criterion.getSlicedOperationForSession(sessionId);
			if (!slicedOp.checkAndSetDataDependency(def)){
				slicedOp.addDef(def);			
				slicedOp.addUses(criterion.getUsesForDef(def));
				ops.add(slicedOp);

				for (Binding<? extends MemoryLocation> binding : criterion.getUsesForDef(def)) {
					BytecodeEvent provider = binding.getProvider();
					ops.addAll(findBackwardSliceForBytecode(provider, 
															binding, 
															sessionId, 
															includeData, 
															includeControl, 
															env));
				}
			}
		}

		//Control flow analysis
		if (includeControl){
			final BytecodeEvent controller = findControllerForBytecode(criterion, env);
			SlicedBytecodeEvent slicedController = controller.getSlicedOperationForSession(sessionId);
			if (!slicedController.checkAndSetControlDependency(def)){
				slicedController.addDefs(controller.getDefs());
				slicedController.addUses(controller.getUses());
				ops.add(slicedController);
				ops.addAll(findBackwardSliceForBytecode(controller, 
														sessionId, 
														includeData, 
														includeControl, 
														env));
			}
		}
		return ops;
	}
	
	private static Set<SlicedBytecodeEvent> findBackwardSliceForBytecode(final BytecodeEvent criterion, 
																			int sessionId,
																			boolean includeData,
																			boolean includeControl, 
																			RuntimeEnv sessionManager){
		Set<SlicedBytecodeEvent> ops = new HashSet<>();
		if (criterion == BytecodeEvent.UNTRACED_BYTECODE_OPERATION){
			return ops;
		}
		
		//Data flow analysis  
		if (includeData){
			SlicedBytecodeEvent slicedOp = criterion.getSlicedOperationForSession(sessionId);
			for (Binding<?> def : criterion.getDefs()){
				if (!slicedOp.checkAndSetDataDependency(def)){
					slicedOp.addDefs(criterion.getDefs());			
					slicedOp.addUses(criterion.getUses());
					ops.add(slicedOp);

					for (Binding<? extends MemoryLocation> binding : criterion.getUses()) {
						BytecodeEvent provider = binding.getProvider();
						ops.addAll(findBackwardSliceForBytecode(provider, 
																binding, 
																sessionId, 
																includeData, 
																includeControl, 
																sessionManager));
					}
				}
			}
		}
		
		
		//Control flow analysis
		if (includeControl){
			final BytecodeEvent controller = findControllerForBytecode(criterion, sessionManager);
			SlicedBytecodeEvent slicedController = controller.getSlicedOperationForSession(sessionId);
			for (Binding<?> def : criterion.getDefs()){
				if (!slicedController.checkAndSetControlDependency(def)){
					slicedController.addDefs(controller.getDefs());
					slicedController.addUses(controller.getUses());
					ops.add(slicedController);
					ops.addAll(findBackwardSliceForBytecode(controller, 
															sessionId, 
															includeData, 
															includeControl, 
															sessionManager));
				}
			}
		}
		return ops;
	}

	public static Set<Integer> findBackwardSliceForBytecodeAsSourceLineNo(final BytecodeEvent criterion, 
																			boolean includeData,
																			boolean includeControl, 
																			RuntimeEnv session){
		List<SlicedBytecodeEvent> events = findBackwardSliceForBytecode(criterion, includeData, includeControl, session);

		Set<Integer> lineNos = new HashSet<>();
		for (SlicedBytecodeEvent slicedBytecodeEvent : events) {
			lineNos.add(slicedBytecodeEvent.getSourceCodeLineNumber());
		}
		return lineNos;
	}
	
	
	/**
	 * This is a full forward slice including both data dependency dependency.
	 * 
	 * @param op
	 * @return
	 */
	public static Set<BytecodeEvent> findDataForwardSliceForBytecode(final BytecodeEvent criterion, 
												 							final boolean includeData, 
												 							final boolean includeControl, 
												 							final RuntimeEnv env){
		Set<BytecodeEvent> forwardSliceSet = new HashSet<>();
		
		forwardSliceSet.add(criterion);
	
		/*
		 	Perform data flow analysis first. For data flow analysis, 
		 	for every value defined by a bytecode, we go forward in
		 	the history to find any usage of that value. The bytecode
		 	using this value is included in the slice. Transitively,
		 	we repeat the same story for the defined values of newly
		 	found bytecode. 
		 */
		
		List<BytecodeEvent> events = env.getAllBytecodeEvents();
		for (Binding<?> def : criterion.getDefs()){
			
			//we skip the pseudo variables in this step 
			if (def.getMemoryLocation() instanceof PseudoVariableMemoryLocation
					)
				continue;
			
			for (long i = criterion.getEventTime(); i < events.size(); i++){
				BytecodeEvent nextEvent = env.getBytecodeEventWithTime(i);
				Binding<?>[] uses = nextEvent.getUses();
				for (Binding<?> use : uses) {
					if (use == def){
						forwardSliceSet.add(nextEvent);
						forwardSliceSet.addAll(findDataForwardSliceForBytecode(nextEvent, true, true, env));
						if (doesDefinePseudoVar(nextEvent)){
							//the next instruction after a CFI is pseudo entry 
							for (long j = i + 2; j < events.size(); j++){
								BytecodeEvent cfd = env.getBytecodeEventWithTime(j);
								if (cfd.isBasicBlockEntry())
									break;
								
								forwardSliceSet.add(cfd);
							}
						}
					}
				}
			}
		}
		
		return forwardSliceSet;
	}

	private static boolean doesDefinePseudoVar(BytecodeEvent e){
		for(Binding<?> def : e.getDefs()){
			if (def.getMemoryLocation() instanceof PseudoVariableMemoryLocation)
				return true;
		}
		
		return false;
	}
	
	
	///////////////////////////////// finding data and control dependency information ///////////////////////////////////
	/**
	 * Given a bytecode operation, this method finds the bytecode operation to
	 * which this bytecode is control dependent on, i.e., the result bytecode 
	 * operation defines the pseudo variable which the argument uses.
	 * 
	 * @param op
	 * @return
	 */
	public static Set<BytecodeEvent> findControllersForBytecode(BytecodeEvent op, RuntimeEnv session){
		if (op.isBasicBlockEntry()){
			throw new IllegalStateException("Trying to find backward slice for a non-bytecode");
		}
		Set<BytecodeEvent> controllers = new HashSet<>();
		BytecodeEvent controller = op;
		do{
			controllers.add(controller);
			controller = findControllerForBytecode(controller, session);
		}while(controller != BytecodeEvent.UNTRACED_BYTECODE_OPERATION);
		
		return controllers;
	}
	
	
	/**
	 * Given a bytecode event, this method find the immediate controller
	 * event that dominates the given event. 
	 * 
	 * includePEI specifies whether this method should consider PEIs as
	 * a control flow instruction or not.  
	 * @param op
	 * @param session
	 * @return
	 */
	public static BytecodeEvent findControllerForBytecode(final BytecodeEvent op, RuntimeEnv session){
		//try to use the cached value
		if (op.getController() != null)
			return op.getController();
		
		//We search for control dependency only among the events 
		//executed within the thread which executed slicing criterion.
		//If the activation record corresponding to the slicing
		//criterion is not traced, we assume the corresponding thread
		//is not available either, so we just return UNTRACED_BC as the 
		//controller
		Activation act = op.getActivationRecord();
		if (act == Activation.UNTRACED_ACTIVATION){
			op.setController(BytecodeEvent.UNTRACED_BYTECODE_OPERATION);
			return BytecodeEvent.UNTRACED_BYTECODE_OPERATION;
		}
		ThreadCallStack tcs = act.getThread();
		
		BytecodeEvent controller = op;
		
		//first we go back in the history to find the basic block to which this op belongs
		while (!controller.isBasicBlockEntry()){
			int index = controller.getOperationIndex() - 1;
			if (index < 0){
				controller = BytecodeEvent.UNTRACED_BYTECODE_OPERATION;
				break;
			}
			controller = tcs.getBytecodeEventAtIndex(index);
		}
		
		// Now we iterate over the basic block entries up to  
		// the point we find a non-basic-block-entry-operation
		// which defines a pseudo variable
		while (controller.isBasicBlockEntry()){
				//There is overlap between PEIs and inherently control 
				//flow instructions e.g.: invocations, throw insn
				//i.e., instructions that affect control flow while
				//they can also throw exception upon execution.
				
			controller = controller.getUses()[0].getProvider();
			if (controller == BytecodeEvent.UNTRACED_BYTECODE_OPERATION)
				break;
		}
		
		op.setController(controller);
		return controller;
	}

	/* *********************************************************************************
	 ******************* Finding a Specific Occurrence of an Event *********************
	 ***********************************************************************************/

	public static BytecodeEvent findLastOccurrenceOfOpCode(int opCode, RuntimeEnv env){
		List<BytecodeEvent> events = env.getAllBytecodeEvents();
		for (int i = events.size() - 1; i>=0; i--){
			if (events.get(i).getOperationCode() == opCode)
				return events.get(i);
		}
		
		throw new IllegalStateException("No event occurred for the specified operation code!");
	}
	
	
	/**
	 * n == 1 means first occurrence,
	 * n == 2 means second occurrence
	 * ...
	 * @param n
	 * @param eventClass
	 * @param session
	 * @return
	 */
	public static BytecodeEvent findNthOccurrenceOfEventClass(int n, int eventClass, RuntimeEnv env){
		int occurrence = 0;
		for (BytecodeEvent event : env.getAllBytecodeEvents()){
			if (event.getEventClassCode() == eventClass){
				occurrence++;
				if (occurrence == n)
					return event;
			}
		}
		
		throw new IllegalStateException("No event occurred for the specified event class!");
	}

	
	/**
	 * n == 1 means last occurrence,
	 * n == 2 means second last occurrence
	 * ...
	 * @param n
	 * @param eventClass
	 * @param session
	 * @return
	 */
	public static BytecodeEvent findNthLastOccurrenceOfEventClass(int n, int eventClass, RuntimeEnv env){
		List<BytecodeEvent> events = env.getAllBytecodeEvents();
		int occurrence = 0;
		for (int i = events.size() - 1; i>=0; i--){
			if (events.get(i).getEventClassCode() == eventClass){
				occurrence++;
				if (occurrence == n)
					return events.get(i);
			}
		}
		
		throw new IllegalStateException("No event occurred for the specified event class!");
	}
	
	/**
	 * n == 1 means last occurrence,
	 * n == 2 means second last occurrence
	 * ...
	 * @param n
	 * @param eventClass
	 * @param session
	 * @return
	 */
	public static BytecodeEvent findNthLastOccurrenceOfOpCode(int n, short opCode, RuntimeEnv env){
		List<BytecodeEvent> events = env.getAllBytecodeEvents();
		int occurrence = 0;
		for (int i = events.size() - 1; i>=0; i--){
			if (events.get(i).getOperationCode() == opCode){
				occurrence++;
				if (occurrence == n)
					return events.get(i);
			}
		}
		
		throw new IllegalStateException("No event occurred for the specified operation code!");
	}

	
	///////////////////////////////////////////////////////////////////////////////
	///////////////////////////Abstract Operations/////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////
	
	
//	private AbstractionSessionManager abstractionSessionManager;
	

	
//	private AbstractSlicedSession fullBytecodeOperationsSession;

	
//	public AbstractionSessionManager getAbstractionSessionManager(){
//	if (this.abstractionSessionManager == null){
//		this.abstractionSessionManager = AbstractionSessionManager.getInstance(this.bytecodes);
//	}
//	
//	return this.abstractionSessionManager;
//}
//
//public BytecodeSlicedRetrievalSession findBackwardSliceForBytecode(int bcIndex){
//	BytecodeEvent criterion = bytecodes.get(bcIndex);
//	int sessionId = nextSessionId ++;
//	return new BytecodeSlicedRetrievalSession(sortAbstractOperations(findBackwardSlice(criterion, sessionId), sessionId), sessionId);
//}

}
